/**
 * @file bipartitegraph.h
 * @author Léo Perrin
 * @date 2011/11/22
 */

#ifndef _GRAPH_H_
#define _GRAPH_H_

#include <vector>
#include <fstream>
#include <string>

#include "vertex.h"

/**
 * @class BipartiteGraph
 * @brief A class to manipulate bipartite graphs.
 *
 * This class models a mathematical bipartite graphs with to sets of
 * vertices, the left ones and the right one. It uses the Vertex class
 * to model its vertices.
 */
class BipartiteGraph
{
	private:
		/**
		 * @brief A std::vector of Vertex instances corresponding to the left
		 * vertices of the graph.
		 */
		std::vector <Vertex> leftVertices;

		/**
		 * @brief A std::vector of Vertex instances corresponding to the left
		 * vertices of the graph.
		 */
		std::vector <Vertex> rightVertices;

	public:
		/**
		 * @brief Uses the given stream input to build a BipartiteGraph
		 * instance.
		 *
		 * The input must be formatted as follow (from the
		 * statement of problem1):
		 * "
		 * m n
		 * 1 : neighbour1
		 * 2 : neighbour1
		 * ...
		 * m : neighbour1
		 * neighbour2 ... neighbourD1
		 * neighbour2 ... neighbourD2
		 * neighbour2 ... neighbourDm
		 * "
		 * where the first line specifies the number of vertices on the
		 * left m and on the right n, and where each following line
		 * specifies a vertex on the left followed by a colon and then
		 * a list of neighbours on the right separated by spaces.
		 */
		BipartiteGraph(std::istream *);

		/**
		 * @brief Returns the number of left vertices in this graph.
		 *
		 * @return The size of leftVertices.
		 */
		unsigned int getNleftVertices();

		/**
		 * @brief Returns the number of right vertices in this graph.
		 * 
		 * @return The size of rightVertices.
		 */
		unsigned int getNrightVertices();
		
		/**
		 * @brief Returns the Vertex of index ind in the set of
		 * the left vertices.
		 *
		 * @param ind The index of the left vertex we want.*/
		Vertex leftVertex(unsigned int);

		/**
		 * @brief Returns the Vertex of index ind in the set of
		 * the left vertices.
		 *
		 * @param ind The index of the right vertex we want.*/
		Vertex rightVertex(unsigned int);
};

#endif
