/**
 * @file vertex.h
 * @author Léo Perrin
 * @date 2011/11/22
 */

#ifndef _VERTEX_H_
#define _VERTEX_H_

#include <vector>

/**
 * @class Vertex
 *
 * @brief Models a mathematical graph's vertex.
 *
 * Stores the index of the vertex as well as the indices of its
 * neighbours, i.e., the indices of the vertices of the "other side"
 * connected to it. We consider bipartite graphs, remember?
 */
class Vertex
{
	private:
		/**
		 * @brief The index of the vertex.
		 */
		unsigned int index;
		
		/**
		 * @brief The indices of the vertices of the other side which
		 * have an edge to this particular vertex.
		 */
		std::vector <unsigned int> neighbours;

	public:

		/**
		 * @brief Constructs a Vertex instance with the given index.
		 *
		 * @param ind The index of the vertex.
		 */
		Vertex(unsigned int);

		/**
		 * @brief Constructs a Vertex instance with the given index and
		 * set of neighbours.
		 *
		 * @param ind The index of the vertex.
		 * @param close A std::vector containing the indices of its
		 * neighbours (from the other side!).
		 */
		Vertex(unsigned int, std::vector<unsigned int>);

		/**
		 * @brief Parses a line of an input file describing a graph.
		 *
		 * When given for instance "2: 3 1 6 7", creates a Vertex with
		 * neighbours 3,1,6,7 and index 2 and puts it in
		 * the leftVertices array.
		 *
		 * @param line The line to parse; it must be formatted like
		 * this: "<vertex index>: <first neighbour index> …", the
		 * indices of the neighbours being separated by spaces.
		 */
		static Vertex line2vertex(std::string);

		/**
		 * @brief Appends the index integer to the list of the neighbours
		 * of the vertex instance.
		 *
		 * @param index The index to add.
		 */
		void addNeighbour(unsigned int);

		/**
		 * @brief Returns the index of the Vertex.
		 *
		 * @return The value of the attribute index.
		 */
		unsigned int getIndex();

		/**
		 * @brief Returns the set of the neighbours of the vertex.
		 *
		 * @return The getNeighbours attribute of the Vertex.
		 */
		std::vector<unsigned int> getNeighbours();

		/**
		 * @brief Returns the order of the vertex, i.e. its number of
		 * neighbours.
		 *
		 * @return The size of the neighbours attribute.
		 */
		unsigned int order();
};

#endif
