/**
 * \file php.cpp
 * \author Léo Perrin
 * \date 2011/11/23
 */
#include <vector>
#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <time.h>

using namespace std;

#include "../cnflib/formula.h"
#include "../cnflib/clause.h"
#include "../cnflib/variableset.h"
#include "../graphlib/bipartitegraph.h"
#include "../graphlib/vertex.h"


/**
 * \brief Transforms a C++ string into an unsigned int using stream
 * buffers.
 *
 * \param str The string to turn into a number.
 *
 * \return The corresponding integer.
 */
unsigned int string2uint(std::string str)
{
	std::istringstream buffer(str);
	unsigned int number;
	buffer >> number;
	return number;
}



/**
 * \brief Adds the clause corresponding to the <em>pigeon axiom</em> for
 * the ith pigeon.
 *
 * \param i The index of the pigeon we consider.
 * \param G The BipartiteGraph instance from which to build the formula.
 * \param form A pointer to the Formula instance we will append the
 * clause to.
 * \param vars The variable set to use.
 */
void pigeonAxiom(unsigned int i,
	BipartiteGraph G,
	Formula * form,
	VariableSet * vars)
{
	vector<int> clauseVar;
	for (unsigned int j=0; j<G.leftVertex(i).order(); j++)
		clauseVar.push_back(  vars->getVarCode\
			(i, G.leftVertex(i).getNeighbours()[j] )  );
	Clause pigeon(clauseVar);
	form->addClause(pigeon);
}




/**
 * \brief Adds the clauses corresponding to the hole <em>axiom</em> for
 * hole number j.
 *
 * \param j The hole to consider.
 * \param G The BipartiteGraph instance from which to build the formula.
 * \param form A pointer to the Formula instance we will append the
 * clause to.
 * \param vars The variable set to use.
 */
void holeAxiom(unsigned int j,
	BipartiteGraph G,
	Formula * form,
	VariableSet * vars)
{
	vector<int> clauseVar;
	for (unsigned int i1=0; i1<G.rightVertex(j).order(); i1++)
		for (unsigned int i2=i1+1; i2<G.rightVertex(j).order(); i2++)
		{
			clauseVar.clear();
			clauseVar.push_back(  (-1)*vars->\
				getVarCode(G.rightVertex(j).getNeighbours()[i1], j)  );
			clauseVar.push_back(  (-1)*vars->\
				getVarCode(G.rightVertex(j).getNeighbours()[i2], j)  );
			Clause hole(clauseVar);
			form->addClause(hole);
		}
}


/**
 * \brief Prints the dimacs formatted CNF formula corresponding to
 * the pigeon hole principle on graph G to the standard output.
 *
 * \param G The BipartiteGraph instance from which to create formula.
 */
void pigeonHolePrinciple(BipartiteGraph G)
{
	string comment ("CNF formula encoding the pigeon principle");
	unsigned int m = G.getNleftVertices();
	unsigned int n = G.getNrightVertices();
	VariableSet vars(m,n);
	Formula form(vars);
	for (unsigned int i=0; i<m; i++)
		pigeonAxiom(i,G,&form,&vars);
	for (unsigned int j=0; j<n; j++)
		holeAxiom(j,G,&form,&vars);
	form.toFile(&cout,comment);
}

/**
 * \brief Prints on standard output the graph file corresponding to a
 * graph with n+1 left vertives, n right vertices and such that all the
 * left vertices have leftOrder neighbours picked uniformly randomly
 * among the right ones.
 *
 * \param leftOrder The number of random distinct neighbours each left
 * vertex must have.
 * \param n The number of left vertices (and the number of right ones
 * minus one).
 */
void randomNeighbours(unsigned int leftOrder,
	unsigned int n)
{
	cout<<n+1<<" "<<n;
	srand(time(NULL));
	for (unsigned int u=1; u<=n+1; u++)
	{
		cout<<endl<<u<<":";
		vector<int> neighbours;
		for (unsigned int i=0; i<leftOrder; i++)
		{
			bool ok=false;
			while (!ok)
			// We make sure we don't pick twice the same vertex
			{
				int rand_index = rand()%n + 1;
				ok = true;
				for (unsigned int j=0; j<neighbours.size(); j++)
					if (rand_index == neighbours[j])
						ok = false;
				if (ok)
					neighbours.push_back(rand_index);
			}
		}
		for (unsigned int i=0; i<neighbours.size()-1; i++)
			cout<<neighbours[i]<<" ";
		cout<<neighbours[neighbours.size()-1];
	}
}

/**
 * \brief Main method to manipulate PHP formulas.
 *
 * If not given any argument, parses stdin to read a graph and outputs
 * the corresponding DIMACS file on standard output. Else, if given 3,
 * outputs on stdout a "graph formatted" text able to generate a
 * complete bipartite graph with m left vertices and n right ones.
 */
int main(int argc, char ** argv)
{
	BipartiteGraph * G;
	if (argc >=4)
	{
		string inArg (argv[1]);
		if (inArg.compare("--generate")==0)
		// Outputs the complete graph for n, m as given in the command
		// line arguments
		{
			unsigned int m = (unsigned int)str2num(string (argv[2]) );
			unsigned int n = (unsigned int)str2num(string (argv[3]) );
			cout<<m<<" "<<n;
			for (unsigned int i=0; i<m; i++)
			{
				cout<<endl<<i+1<<":";
				for (unsigned int j=0; j<n; j++)
					cout<<j+1<<" ";
			}
		}
		else if (inArg.compare("--generate-random")==0)
		// Outputs the random graph with n vertices and of left degree d
		// as given in the command line arguments
		{
			unsigned int n = (unsigned int)str2num(string (argv[2]) );
			unsigned int d = (unsigned int)str2num(string (argv[3]) );
			randomNeighbours(d,n);
		}
	}
	else if (argc >= 2)
	// reading graph from file
	{
		string inArg (argv[1]);
		ifstream fileIn(inArg.c_str());
		BipartiteGraph G(&fileIn);
		pigeonHolePrinciple(G);
	}
	else
	// reading graph from standard inArg
	{
		BipartiteGraph G(&cin);
		pigeonHolePrinciple(G);
	}
	cout<<endl;

	return 0;
}
