#include <stdlib.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <string>

#include "../libcnf/libcnf.h"

using namespace cnf;


/**
 * @def NONE
 * The value to put in a grid when the value of a cell is not
 * known. It ought to be different from 0..9.
 */
#define NONE 100



/**
 * @brief Adds the clauses encoding the fact that each cell must have
 * one and only one value to the Formula form.
 *
 * First, we add \f$\bar{x_{ijk}} \land \bar{x_{ijk'}}\f$ for k<k'
 * and then, second, we add \f$\lor_{k=0,9} x_{ijk}\f$
 */
void oneValuePerCell(
	unsigned int i, unsigned int j, // coordinates of the cell
	Formula * form) // formula to which the clauses will be added
{
	// First, we add \f$\bar{x_ijk} \land \bar{x_ijk'}\f$ for k<k'
	for (unsigned int k1=0; k1<8; k1++)
		for (unsigned int k2=k1+1; k2<9; k2++)
		{
                        form->newClause();
			form->addLit(0,false,i,j,k1);
			form->addLit(0,false,i,j,k2);
			form->pushClause();
		}
	// Second, we add \lor_{k=0,9} x_ijk
	form->newClause();
	for (unsigned int k1=0; k1<9; k1++)
		form->addLit(0,true,i,j,k1);
	form->pushClause();
}



/**
 * @brief Adds the clauses corresponding to the uniqueness of each
 * number in a line.
 *
 * We add \f$\bar{x_{ijk}} \land \bar{x_{ij'k}}\f$ for j<j', for all
 * k in the Formula form.
 */
void oneValuePerLine(
	unsigned int i, // number of the line
	Formula * form) // formula to which the clauses will be added
{
	for (unsigned int k=0; k<9; k++)
		for (unsigned int j1=0; j1<8; j1++)
			for (unsigned int j2=j1+1; j2<9; j2++)
			{
				form->newClause();
				form->addLit(0,false,i,j1,k);
				form->addLit(0,false,i,j2,k);
				form->pushClause();
			}
}

/**
 * @brief Adds the clauses corresponding to the uniqueness of each
 * number in a column.
 *
 * We add \f$\bar{x_{ijk}} \land \bar{x_{i'jk}}\f$ for i<i', for all
 * k to the Formula form.
 */
void oneValuePerColumn(
	unsigned int j, // number of the column
	Formula * form) // formula to which the clauses will be added
{
	for (unsigned int k=0; k<9; k++)
		for (unsigned int i1=0; i1<8; i1++)
			for (unsigned int i2=i1+1; i2<9; i2++)
			{
				form->newClause();
				form->addLit(0,false,i1,j,k);
				form->addLit(0,false,i2,j,k);
				form->pushClause();
			}
}


/**
 * @brief Adds the clauses corresponding to the fact that each number
 * appears only once in each square.
 *
 * Adds the clauses \f$\bar{x_{ijk}} \land \bar{x_{i'j'k}}\f$ for
 * all (i,j) and (i',j') in the square and for all k in [0,8].
 */
void oneValuePerSquare(
	unsigned int i, // i=0,2 : "line" of the square to consider
	unsigned int j, // j=0,2 : "column" of the square to consider
	Formula * form) // formula to which the clauses will be added
{
	for (unsigned int k=0; k<9; k++)
		for (unsigned int l1=0; l1<8; l1++)
			for (unsigned int l2=l1+1; l2<9; l2++)
			{
				form->newClause();
				form->addLit(0,false, i*3+l1/3, j*3+l1%3, k);
				form->addLit(0,false, i*3+l2/3, j*3+l2%3, k);
				form->pushClause();
			}
}


/**
 * @brief Adds the clauses corresponding to an assignment of the
 * literals known for the sudoku grid.
 *
 * For each cell of grid, if it is known, adds a one-literal clause to
 * the formula form for each literal corresponding to it.
 */
void grid2restriction(std::vector< std::vector<int> > grid, Formula * form, VariableSet * vars)
{
	for (int i=0; i<9; i++)
		for (int j=0; j<9; j++)
		{
			if (grid[i][j] != NONE)
				for (int k=0; k<9; k++)
					if (k==grid[i][j]-1)
					{
						form->newClause();
						form->addLit(0,true,i,j,k);
						form->pushClause();
					}
		}
}

/**
 * @brief Reads a sudoku grid from the input stream and stores it in a
 * std::vector of std::vector of int.
 *
 * @return A std::vector of std::vector of int containing the sudoku grid with
 * the NONE value where the number is not known.
 */
std::vector< std::vector<int> > parseGrid(std::istream * input)
{
	std::vector< std::vector<int> > grid;
	for (unsigned int i=0; i<9; i++)
	{
		std::vector<int> line;
 //		std::cout<<std::endl;
		for (unsigned int j=0; j<9; j++)
		{
			char n;
			(*input) >> n;
			if ((*input).good())
			{
//				std::cout << n ;
				if (n!='-')
					line.push_back(n-'0');
				else
					line.push_back(NONE);
			}
			else
			//if the stream is stopped
			{
				std::cout<<"error 403: input stream does not contain enough"
				    <<"information !"<<std::endl;
				exit(403);
			}
		}
		grid.push_back(line);
	}
	std::cout<<"Finish parsing grid"<<std::endl;
	return grid;
}


int main(int argc, char ** argv)
{
	VariableSet vars(9,9,9);
	Formula form(&vars);
	// Exactly one value per cell
	for (unsigned int i=0; i<9; i++)
		for (unsigned int j=0; j<9; j++)
			oneValuePerCell(i,j,&form);
	// Different element on each line
	for (unsigned int i=0; i<9; i++)
		oneValuePerLine(i,&form);
	// Different element on each column
	for (unsigned int j=0; j<9; j++)
		oneValuePerColumn(j,&form);
	// Different element in each "big square"
	for (unsigned int i=0; i<3; i++)
		for (unsigned int j=0; j<3; j++)
			oneValuePerSquare(i,j,&form);
	// Adding the clauses corresponding to fixed variables (i.e, depends
	// on the sudoku grid, so it's time to read it).
	std::string comment ("CNF formula corresponding to the sudoku");
	if (argc==1)
	{
		grid2restriction(parseGrid(&std::cin),&form,&vars);
		form.toFile(&std::cout,comment);
	}
	else
	{
		std::string name (argv[1]);
		std::ifstream input(name.c_str());
		std::cout<<"Reading file "<<name<<std::endl;
		grid2restriction(parseGrid(&input),&form,&vars);
		input.close();
		name+=".dimacs";
		std::ofstream output(name.c_str());
		form.toFile(&output,comment);
		std::cout<<std::endl<<"DIMACS file written: "<<name<<std::endl;
	}
	std::cout<<std::endl;
	return 0;
}
