/**
 * @file   slitherlink.h
 * @author Léo Perrin
 * @date   Time-stamp: <2012-05-18 16:50:01 leo>
 * 
 * @brief Contains the header of the Slitherlink class.
 */
#ifndef _SLITHERLINK_H_
#define _SLITHERLINK_H_


#include <vector>
#include <iostream>

/**
 * @class Slitherlink
 *
 * @brief Models a slitherlink puzzle and allows an easy manipulation
 * of it (input, output, etc).
 *
 * See <a
 * href="http://en.wikipedia.org/wiki/Slitherlink">wikipedia</a> for a
 * detailed description of this puzzle.
 */
class Slitherlink
{
private:
        /**
         * @brief The horizontal vertices of the grid take value 1 if
         * they are part of the loop, 0 otherwise.
         *
         * It has one more line than the hints grid.
         */
        std::vector<std::vector<bool> > hVertices;

        /**
         * @brief The vertical vertices of the grid take value 1 if
         * they are part of the loop, 0 otherwise.
         *
         * It has one more column than the hints grid.
         */
        std::vector<std::vector<bool> > vVertices;
        
        /**
         * @brief The width of the grid containing the hints.
         */
        unsigned int width;

        /**
         * @brief The height of the grid containing the hints.
         * 
         * For instance if the grid with the hints contains 4 lines,
         * then height=5.
         */
        unsigned int height;

        /**
         * @brief The hints contained in the grid.
         *
         * If no hints where given for a cell, then it will contain
         * -1.
         */
        std::vector<std::vector<int> > hints;


public:
        /**
         * @brief Initialises all the attributes to empty/zero values.
         */
        Slitherlink();

        /**
         * @brief Reads a slitherlink grid from the input and assigns
         * all the attributes accordingly.
         * 
         * @param input The stream where the information must be read.
         */
        void parseInput(std::istream *input);

        /**
         * @brief Assigns a new value to the vertex at coordinate
         * (i,j). The horizontal vertex is assigned if the boolean is
         * true, the vertical one is otherwise.
         * 
         * @param horizontal Which type of vertex should we consider.
         * @param i          The index of the line of the node.
         * @param j          The index of the column of the node.
         * @param value      The value to give to the node.
         */
        void setVertex(bool horizontal, unsigned int i,
                       unsigned int j, bool value);

        /**
         * @brief Returns the hint of coordinates (i,j). If there is
         * no hint at this point, returns -1.
         * 
         * @param i The index of the line.
         * @param j The index of the column.
         * 
         * @return The value of the hint.
         */
        int getHint(unsigned int i, unsigned int j);

        unsigned int getWidth();
        unsigned int getHeight();

        /**
         * @brief Prints on standard output the line on top of the one
         * containing the hints at line i.
         *
         * Deals with both horizontal vertices and the vertical ones
         * that cross it.
         * 
         * @param i The line we want to print.
         */
        void printNonHintLine(unsigned int i);

        /**
         * @brief Prints on standard output the solution of the grid
         * according to the content of the nodes attributes.
         */
        void printSolution();
};


#endif
