/**
 * @file   slitherlink2formula.cpp
 * @author Léo Perrin
 * @date   Time-stamp: <2012-06-19 16:55:12 leo>
 * 
 * @brief Contains the functions used to turn a Slitherlink puzzle into
 * a CNF formula.
 */

#include <stdlib.h>

#include "../libcnf/libcnf.h"
#include "slitherlink.h"



#define HORIZONTAL 0
#define VERTICAL 1
#define HORIZONTAL_EDGE_INDEX 2
#define   VERTICAL_EDGE_INDEX 3
#define HORIZONTAL_EDGE_SENSE 4
#define   VERTICAL_EDGE_SENSE 5

#define NUMBERING 2
#define SENSE 4


/**
 * @brief Rather self explanatory: it is used to easily store and
 * retrieve coordinates in the grid.
 */
struct coord
{
        /**
         * Equals HORIZONTAL if horizontal, VERTICAL if vertical.
         */
        int orientation;
        unsigned int x;
        unsigned int y;
};



// ===================================================================
// ================================================= Continuity axioms
// ===================================================================



/**
 * @brief Retrieves the codes of the neighbours of a given horizontal
 * edge depending on its position in the grid.
 * @param[out] neighbours1 When the call is done, contains the
 *             coordinates of the variables on the right of the vertex.
 * @param[out] neighbours2 When the call is done, contains the
 *             coordinates of the variables on the left of the vertex.
 */
void getHneighbours(Slitherlink * sl,
                    unsigned int i, unsigned int j,
                    std::vector<coord> * neighbours1,
                    std::vector<coord> * neighbours2 )
{
        neighbours1->clear();
        neighbours2->clear();
        if (i > 0) {
                coord c;
                c.orientation = VERTICAL;
                c.x = i-1; c.y = j+1;
                neighbours1->push_back(c);
        } // 1
        if (j < sl->getWidth()-1) {
                coord c;
                c.orientation = HORIZONTAL;
                c.x = i; c.y = j+1;
                neighbours1->push_back(c);
        } // 2
        if (i < sl->getHeight()) {
                coord c;
                c.orientation = VERTICAL;
                c.x = i; c.y = j+1;
                neighbours1->push_back(c);
        } // 3
        if (i < sl->getHeight()) {
                coord c;
                c.orientation = VERTICAL;
                c.x = i; c.y = j;
                neighbours2->push_back(c);
        } // 4
        if (j > 0) {
                coord c;
                c.orientation = HORIZONTAL;
                c.x = i; c.y = j-1;
                neighbours2->push_back(c);
        } // 5
        if (i > 0) {
                coord c;
                c.orientation = VERTICAL;
                c.x = i-1; c.y = j;
                neighbours2->push_back(c);
        } // 6
}


/**
 * @brief Retrieves the codes of the neighbours of a given vertical
 * edge depending on its position in the grid.
 * @param[out] neighbours1 When the call is done, contains the
 *             coordinates of the variables on top of the vertex.
 * @param[out] neighbours2 When the call is done, contains the
 *             coordinates of the variables below the vertex.
 */
void getVneighbours(Slitherlink * sl,
                    unsigned int i, unsigned int j,
                    std::vector<coord> * neighbours1,
                    std::vector<coord> * neighbours2 )
{
        neighbours1->clear();
        neighbours2->clear();
        if (j > 0) {
                coord c;
                c.orientation = HORIZONTAL;
                c.x = i; c.y = j-1;
                neighbours1->push_back(c);
        } // 1
        if (i > 0) {
                coord c;
                c.orientation = VERTICAL;
                c.x = i-1; c.y = j;
                neighbours1->push_back(c);
        } // 2
        if (j < sl->getWidth()) {
                coord c;
                c.orientation = HORIZONTAL;
                c.x = i; c.y = j;
                neighbours1->push_back(c);
        } // 3
        if ((i < sl->getHeight()) && (j < sl->getWidth())) {
                coord c;
                c.orientation = HORIZONTAL;
                c.x = i+1; c.y = j;
                neighbours2->push_back(c);
        } // 4
        if (i < sl->getHeight()-1) {
                coord c;
                c.orientation = VERTICAL;
                c.x = i+1; c.y = j;
                neighbours2->push_back(c);
        } // 5
        if (i < sl->getHeight() && j>0) {
                coord c;
                c.orientation = HORIZONTAL;
                c.x = i+1; c.y = j-1;
                neighbours2->push_back(c);
        } // 6
}


/** 
 * @brief Writes axioms enforcing the loop to be continuous, i.e. that
 * if an edge is in the loop then exactly one of its neighbous is.
 *
 * The idea is to run this on the two sets of neighbours of a given
 * edge. It will add the following axioms:
 * 
 * @param form 
 * @param edgeCode 
 * @param neighbours 
 */
void addContinuityAxioms(cnf::Formula * form, int edgeCode,
                         std::vector<coord> neighbours)
{
        // At least one of the neighbours is true if the edge is
        form->newClause();
        form->addLitCode((-1)*edgeCode);
        for (unsigned int k=0; k<neighbours.size(); k++)
                form->addLit(neighbours[k].orientation, true,
                             neighbours[k].x, neighbours[k].y);
        form->pushClause();
        // They are mutually exclusive
        if (neighbours.size() > 1)
                for (unsigned int k=0; k<neighbours.size()-1; k++)
                        for (unsigned int l=k+1; l<neighbours.size(); l++)
                        {
                                form->newClause();
                                form->addLitCode((-1)*edgeCode);
                                form->addLit(neighbours[k].orientation,false,
                                             neighbours[k].x, neighbours[k].y);
                                form->addLit(neighbours[l].orientation,false,
                                             neighbours[l].x, neighbours[l].y);
                                form->pushClause();
                        }
}


// ==================================================================
// ======================================================= Hint axioms
// ==================================================================


/** 
 * @brief Returns a vector containing the code of the positive
 * literals corresponding to the $x_e$ variables for each $e$ around
 * the cell at $(i,j)$.
 * 
 * @param vars The variable set.
 * @param i    The first coordinate of the cell.
 * @param j    The second one.
 * 
 * @return The corresponding variable codes.
 */
std::vector<int> hintNeighbours(cnf::VariableSet * vars,
                                unsigned int i, unsigned int j)
{
        std::vector<int> neighbours(4);
        neighbours[0] = vars->getVarCode(HORIZONTAL,i,j);   // 1
        neighbours[1] = vars->getVarCode(  VERTICAL,i,j+1); // 2
        neighbours[2] = vars->getVarCode(HORIZONTAL,i+1,j); // 3
        neighbours[3] = vars->getVarCode(  VERTICAL,i,j);   // 4
        return neighbours;
}


/** 
 * @brief Add the clauses corresponding to a hint equal to zero,
 * i.e. that all edges around it should be set to false.
 */
void addHintAxioms0(cnf::Formula * form, Slitherlink * sl,
                   cnf::VariableSet * vars, unsigned int i,
                   unsigned int j)
{
        std::vector<int> neighbours = hintNeighbours(vars,i,j);
        for (unsigned int k=0; k<4; k++)
        {
                form->newClause();
                form->addLitCode((-1)*neighbours[k]);
                form->pushClause();
        }
}


/** 
 * @brief Add the clauses corresponding to a hint equal to 1, i.e. the
 * following set of clauses:
 * - at least one of the edge is true.
 * - they are mutually exclusive.
 */
void addHintAxioms1(cnf::Formula * form, Slitherlink * sl,
                   cnf::VariableSet * vars, unsigned int i,
                   unsigned int j)
{
        std::vector<int> neighbours = hintNeighbours(vars,i,j);
        // at least one is true
        form->newClause();
        for (unsigned int k=0; k<4; k++)
                form->addLitCode(neighbours[k]);
        form->pushClause();
        // they are mutually exclusive
        for (unsigned int k=0; k<3; k++)
                for (unsigned int l=k+1; l<4; l++)
                {
                        form->newClause();
                        form->addLitCode((-1)*neighbours[k]);
                        form->addLitCode((-1)*neighbours[l]);
                        form->pushClause();
                }
}


/** 
 * @brief Adds the clauses corresponding to a hint being equal to 2,
 * namely these:
 * - All 3 tuples among the 4 variables are true
 * - So do all the negative 3 tuples.
 */
void addHintAxioms2(cnf::Formula * form, Slitherlink * sl,
                   cnf::VariableSet * vars, unsigned int i,
                   unsigned int j)
{
        std::vector<int> neighbours = hintNeighbours(vars,i,j);
        // every 3-tuple of positive literals is true and so is
        // every 3-tuple of negative literals.
        for (unsigned int k=0; k<2; k++)
                for (unsigned int l=k+1; l<3; l++)
                        for (unsigned int m=l+1; m<4; m++)
                        {
                                // positive 3-tuple
                                form->newClause();
                                form->addLitCode(neighbours[k]);
                                form->addLitCode(neighbours[l]);
                                form->addLitCode(neighbours[m]);
                                form->pushClause();
                                // negative 3-tuple
                                form->newClause();
                                form->addLitCode((-1)*neighbours[k]);
                                form->addLitCode((-1)*neighbours[l]);
                                form->addLitCode((-1)*neighbours[m]);
                                form->pushClause();
                        }
}


/** 
 * @brief Adds clauses corresponding to a hint being equal to 3, namely:
 * - at least one of the edges is set to false
 * - the negativity of the edges are mutually exclusive
 */
void addHintAxioms3(cnf::Formula * form, Slitherlink * sl,
                   cnf::VariableSet * vars, unsigned int i,
                   unsigned int j)
{
        std::vector<int> neighbours = hintNeighbours(vars,i,j);
        // At least one is false
        form->newClause();
        for (unsigned int k=0; k<4; k++)
                form->addLitCode((-1)*neighbours[k]);
        form->pushClause();
        // In every subset of size two, at least one is true
        for (unsigned int k=0; k<3; k++)
                for (unsigned int l=k+1; l<4; l++)
                {
                        form->newClause();
                        form->addLitCode(neighbours[k]);
                        form->addLitCode(neighbours[l]);
                        form->pushClause();
                }        
}


void addHintAxioms(cnf::Formula * form, Slitherlink * sl,
                   cnf::VariableSet * vars, unsigned int i,
                   unsigned int j)
{
        switch (sl->getHint(i,j))
        {
        case (-1): break;
        case (0): { addHintAxioms0(form,sl,vars,i,j); break; } 
        case (1): { addHintAxioms1(form,sl,vars,i,j); break; }
        case (2): { addHintAxioms2(form,sl,vars,i,j); break; }
        case (3): { addHintAxioms3(form,sl,vars,i,j); break; }
        }
}



// ===================================================================
// ================================================ Single loop axioms
// ===================================================================

/** 
 * @brief Adds axioms stating that if an edge is activated then it has
 * a unique index.

 * @param form        The formula where to put the clauses
 * @param orientation The orientation of the edge (HORIZONTAL or not?)
 * @param i           Abscissa of the edge
 * @param j           Ordena of the edge
 * @param maxLoopSize The maximum index it can have
 */
void addContinuityAndNumberingAxioms(cnf::Formula * form, int orientation,
                                     unsigned int i, unsigned int j,
                                     unsigned int maxLoopSize)
                                     
{
        // If the edge is in the loop then it has to have an index
        form->newClause();
        form->addLit(orientation,false,i,j);
        for (unsigned int k=0; k<maxLoopSize; k++)
                form->addLit(orientation+NUMBERING,true,i,j,k);
        form->pushClause();
        // If the edge has an index then it has to be in the loop
        for (unsigned int k=0; k<maxLoopSize; k++)
        {
                form->newClause();
                form->addLit(orientation+NUMBERING,false,i,j,k);
                form->addLit(orientation,true,i,j);
                form->pushClause();
        }
}


/** 
 * @brief Adds axioms stating that the if the edge at $(i,j)$ with
 * orientation $orientation$ has index $index$ then:
 *
 * - if $neighbours$ is the neighbours being greater than $edge$
 * according to the variable $sense$, then there must be an edge in
 * $neighbours$ with index equal to $index+1$ or 0.  - otherwise and
 * if $index$ is not zero there must an edge in $neighbours$ with
 * index $index-1$.
 */
void addIncreasingIndexAxioms(cnf::Formula * form, int orientation,
                              unsigned int i, unsigned int j,
                              unsigned int index,
                              bool isNeighboursGreater, bool sense,
                              std::vector<coord> neighbours)
{
        if ((index != 0) || (isNeighboursGreater))
        {
                unsigned int neighboursIndex =
                        (isNeighboursGreater) ? index+1 : index-1;
                // The neighbours must have the correct index
                form->newClause();
                form->addLit(orientation+NUMBERING,false,i,j,index);
                form->addLit(orientation+SENSE,!sense,i,j);
                for (unsigned int k=0; k<neighbours.size(); k++)
                {
                        form->addLit(neighbours[k].orientation+NUMBERING,true,
                                     neighbours[k].x,neighbours[k].y,neighboursIndex);
                        if (isNeighboursGreater) // then maybe the next index is zero (closing the loop)
                                form->addLit(neighbours[k].orientation+NUMBERING,true,
                                             neighbours[k].x,neighbours[k].y,0);
                }
                form->pushClause();
        }
        // else: nothing to do here *takes his jetpack*
}


/** 
 * @brief It, well, places the zero obviously. It adds clauses to the
 * formula enforcing that the edge with index zero has to touch a
 * non-zero hint (the first one encountered) and that there can only
 * be one around this hint. It then sets all the other edges to be
 * non-zero.
 */
void placeZero(cnf::Formula * form, Slitherlink * sl,
               cnf::VariableSet * vars,
               unsigned int i, unsigned int j)
{
        std::vector<int> neighbours(4);
        neighbours[0] = vars->getVarCode(HORIZONTAL_EDGE_INDEX,i  ,j  ,0);
        neighbours[1] = vars->getVarCode(  VERTICAL_EDGE_INDEX,i  ,j+1,0);
        neighbours[2] = vars->getVarCode(HORIZONTAL_EDGE_INDEX,i+1,j  ,0);
        neighbours[3] = vars->getVarCode(  VERTICAL_EDGE_INDEX,i  ,j  ,0);
        // zero is along the side of this cell ...
        form->newClause();
        form->addLitCode(neighbours[0]);
        form->addLitCode(neighbours[1]);
        form->addLitCode(neighbours[2]);
        form->addLitCode(neighbours[3]);
        form->pushClause();
        // ... and there is only one ...
        for (unsigned int k=0; k<3; k++)
                for (unsigned int l=k+1; l<4; l++)
                {
                        form->newClause();
                        form->addLitCode((-1)*neighbours[k]);
                        form->addLitCode((-1)*neighbours[l]);
                        form->pushClause();
                }
        // ... and it can't be anywhere else
        for (unsigned int k=0; k<=sl->getHeight(); k++)
                for (unsigned int l=0; l<sl->getWidth(); l++)
                        if ((l!=j) || ((k!=i) && (k!=i+1)) )
                        {
                                form->newClause();
                                form->addLit(HORIZONTAL_EDGE_INDEX,false,k,l,0);
                                form->pushClause();
                        }
        for (unsigned int k=0; k<sl->getHeight(); k++)
                for (unsigned int l=0; l<=sl->getWidth(); l++)
                        if ((k!=i) || ((l!=j) && (l!=j+1)) )
                        {
                                form->newClause();
                                form->addLit(VERTICAL_EDGE_INDEX,false,k,l,0);
                                form->pushClause();
                        }
}

// ===================================================================
// ===================================================== Main function
// ===================================================================


int main(int argc, char * argv[])
{
        // initialising input and output streams
        if (argc <= 2)
        {
                std::cout<<"Wrong number of argument!"<<std::endl
                         <<"./i2f <slitherlink to solve> <where to put "
                         <<"the dimacs>"<<std::endl;
                exit(1);
        }
        std::ifstream * input  = new std::ifstream(argv[1]);
        std::ofstream * output = new std::ofstream(argv[2]);

        // reading slitherlink
        Slitherlink * sl = new Slitherlink();        
        sl->parseInput(input);
        free(input);

        // ---- creating the variables
        unsigned int maxLoopSize = sl->getHeight()*sl->getWidth()*4;
        // horizontal edges continuity
        cnf::VariableSet * vars = new
                cnf::VariableSet(sl->getHeight()+1,sl->getWidth());
        // vertical edges continuity
        vars->addVars(sl->getHeight(),sl->getWidth()+1);
        // horizontal edges indices
        vars->addVars(sl->getHeight()+1,sl->getWidth(),maxLoopSize+1);
        // vertical edges continuity
        vars->addVars(sl->getHeight(),sl->getWidth()+1,maxLoopSize+1);
        // horizontal edges sense
        vars->addVars(sl->getHeight()+1,sl->getWidth());
        // vertical edges sense
        vars->addVars(sl->getHeight(),sl->getWidth()+1);
        
        // creating the formula corresponding to the slitherlink
        cnf::Formula* form = new cnf::Formula(vars,output,
                                              "formula modeling a slitherlink");

        // several variables which will turn out to be useful
        std::vector<coord> neighbours1;
        std::vector<coord> neighbours2;
        int edgeCode;
        // taking care of horizontal edges
        for (unsigned int i=0; i <= sl->getHeight(); i++)
                for (unsigned int j=0; j < sl->getWidth(); j++)
                {
                        edgeCode = vars->getVarCode(0,i,j);
                        getHneighbours(sl,i,j, &neighbours1,&neighbours2);
                        addContinuityAxioms(form,edgeCode,neighbours1);
                        addContinuityAxioms(form,edgeCode,neighbours2);
                        addContinuityAndNumberingAxioms(form,HORIZONTAL,i,j,
                                                        maxLoopSize);
                        for (unsigned int index=0; index<maxLoopSize; index++)
                        {
                                // positive sense
                                addIncreasingIndexAxioms(form,HORIZONTAL,i,j,
                                                         index, false,true,
                                                         neighbours1);
                                addIncreasingIndexAxioms(form,HORIZONTAL,i,j,
                                                         index, true,true,
                                                         neighbours2);
                                // negative sense
                                addIncreasingIndexAxioms(form,HORIZONTAL,i,j,
                                                         index, true,false,
                                                         neighbours1);
                                addIncreasingIndexAxioms(form,HORIZONTAL,i,j,
                                                         index, false,false,
                                                         neighbours2);
                        }
                }                        
        // taking care of vertical edges
        for (unsigned int i=0; i < sl->getHeight(); i++)
                for (unsigned int j=0; j <= sl->getWidth(); j++)
                {
                        edgeCode = vars->getVarCode(1,i,j);
                        getVneighbours(sl,i,j, &neighbours1,&neighbours2);
                        addContinuityAxioms(form,edgeCode,neighbours1);
                        addContinuityAxioms(form,edgeCode,neighbours2);
                        addContinuityAndNumberingAxioms(form,VERTICAL,i,j,
                                                        maxLoopSize);
                        for (unsigned int index=0; index<maxLoopSize; index++)
                        {
                                // positive sense
                                addIncreasingIndexAxioms(form,VERTICAL,i,j,
                                                         index, false,true,
                                                         neighbours1);
                                addIncreasingIndexAxioms(form,VERTICAL,i,j,
                                                         index, true ,true,
                                                         neighbours2);
                                // negative sense
                                addIncreasingIndexAxioms(form,VERTICAL,i,j,
                                                         index, true ,false,
                                                         neighbours1);
                                addIncreasingIndexAxioms(form,VERTICAL,i,j,
                                                         index, false,false,
                                                         neighbours2);
                        }
                }
        // Adding hint axioms and forcing zero to be around a hint
        bool zeroHasBeenPut = false;
        for (unsigned int i=0; i<sl->getHeight(); i++)
                for (unsigned int j=0; j<sl->getWidth(); j++)
                {
                        addHintAxioms(form,sl,vars,i,j);
                        if ((!zeroHasBeenPut) && (sl->getHint(i,j)>0))
                        {
                                placeZero(form,sl,vars,i,j);
                                zeroHasBeenPut = true;
                        }
                }
                                

        // At last, outputing the eagerly awaited formula
        form->flushCNF();
        // Everything went as expected (or not, hahahaha) so returns 0
        return 0;
}
